/*-
 * Copyright (C)2012,2013 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2012,2013 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: 2sw.c,v 1.8 2013/05/19 19:20:56 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <strings.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#ifdef __FreeBSD__
# include <babolo/BLINflag.h>
#else
# include "linux.h"
#endif

#include "transposit.h"

/*****************************************************************************************************
 *     net  2x2                                                  *
 *  nin   nou .                                                                     *
 *****************************************************************************************************/
transposit_2sw *
transposit_2swnew(u_int32_t flags, size_t in, size_t out, size_t net) {
    transposit_2sw *sw;

    if  (!(sw = calloc(1, sizeof(transposit_link) * out + sizeof(transposit_2sw)))) {
        errno = ENOMEM;
        goto out;
    }
    if  (!(sw->net = calloc(net, sizeof(transposit_2swnet)))) {
        errno = ENOMEM;
        goto out;
    }
    sw->flags = flags;
    sw->nin = in;
    sw->nou = out;
    sw->num = net;
    if  (0) {
out:    if  (sw) {
            if  (sw->net) free(sw->net);
            free(sw);
        }
        sw = NULL;
    }
    return(sw);
}

/*****************************************************************************************************
 *    transposit_2sw                                                              *
 *****************************************************************************************************/
void
transposit_2swfree(transposit_2sw *sw) {
    if  (sw) {
        if  (sw->net) free(sw->net);
        free(sw);
}   }

/*****************************************************************************************************
 *    2x2 sw.                                                                *
 *    oulnk  ousw   inlnk  insw .           *
 * TRANSPOSIT_EXT   ousw  insw      .            *
 *****************************************************************************************************/
int
transposit_2swconn( transposit_2sw *sw
                  , u_int32_t ousw
                  , u_int16_t oulnk
                  , u_int32_t insw
                  , u_int16_t inlnk
                  ) {
    int ex = EX_OK;

    if  (ousw == TRANSPOSIT_EXT) {
        if  (oulnk >= sw->nou) {
            ex = EX_DATAERR;
            errno = EDOM;
            goto out;
        }
    } else if (oulnk >= 2) {
        ex = EX_DATAERR;
        errno = EDOM;
        goto out;
    } else if (ousw >= sw->num) {
        ex = EX_DATAERR;
        errno = EDOM;
        goto out;
    }
    if  (insw == TRANSPOSIT_EXT) {
        if  (inlnk >= sw->nin) {
            ex = EX_DATAERR;
            errno = EDOM;
            goto out;
        }
        sw->out[inlnk].prev = ousw;
        sw->out[inlnk].lnkn = oulnk;
    } else {
        if (inlnk >= 2) {
            ex = EX_DATAERR;
            errno = EDOM;
            goto out;
        } else if (insw >= sw->num) {
            ex = EX_DATAERR;
            errno = EDOM;
            goto out;
        }
        sw->net[insw].src[inlnk].prev = ousw;
        sw->net[insw].src[inlnk].lnkn = oulnk;
    }
out:
    return(ex);
}

/*****************************************************************************************************
 *                                                                         *
 *****************************************************************************************************/
transposit_2sw *
transposit_2swrevers(transposit_2sw *sw) {
    transposit_2sw *sv;
    size_t i, k;

    if  (!(sv = calloc(1, sizeof(transposit_link) * sw->nin + sizeof(transposit_2sw)))) {
        errno = ENOMEM;
        goto out;
    }
    if  (!(sv->net = calloc(sw->num, sizeof(transposit_2swnet)))) {
        errno = ENOMEM;
        goto out;
    }
    sv->flags = sw->flags;
    sv->nin = sw->nou;
    sv->nou = sw->nin;
    sv->num = sw->num;
    for (i = 0; i < sw->num; ++i) {
        sv->net[i].state = sw->net[i].state;
        for (k = 0; k < 2; ++k) {
            if  (transposit_2swconn(sv, i, k, sw->net[i].src[k].prev, sw->net[i].src[k].lnkn)) {
                goto out;
    }   }   }
    for (k = 0; k < sw->nou; ++k) {
        if  (transposit_2swconn(sv, TRANSPOSIT_EXT, k, sw->out[k].prev, sw->out[k].lnkn)) goto out;
    }
    if  (0) {
out:    if  (sv) {
            if  (sv->net) free(sv->net);
            free(sv);
        }
        sv = NULL;
    }
    return(sv);
}

/*****************************************************************************************************
 *    2x2   sw.                                                    *
 *  bits (     )    .       *
 *****************************************************************************************************/
int
transposit_2swnetset(transposit_2sw *sw, void *bits) {
    int ex = EX_OK;
    u_char *bit;
    size_t i;
    u_int k;

    for (i = 0, bit = bits; i < sw->num; ) {
        for (k = 1; k < 0xFF; k <<= 1) {
            sw->net[i++].state = (*bit & k) ? 1 : 0;
            if  (i >= sw->num) break;
        }
        ++bit;
    }
    return(ex);
}

/*****************************************************************************************************
 *         2x2 sw.                   *
 *   swe,    ,                                                     *
 *  *swe != NULL,       .                 *
 *    transposit_switch   free(transposit_switch *).               *
 *****************************************************************************************************/
transposit_switch *
transposit_2swget(transposit_2sw *sw, transposit_switch **swe) {
    transposit_switch *swr = NULL;
    size_t i, l, f = 0;
    transposit_link s;
    int ex = EX_OK;

    if  (!swe) {
        if  (!(swr = calloc(1, sizeof(size_t) * sw->nou + sizeof(transposit_switch)))) {
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
        }
        f = 1;
    } else if (!*swe) {
        if  (!(swr = calloc(1, sizeof(size_t) * sw->nou + sizeof(transposit_switch)))) {
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
        }
        *swe = swr;
        f = 1;
    } else {
        swr = *swe;
    }
    swr->nin = sw->nin;
    swr->nou = sw->nou;
    swr->flags = sw->flags;
    for (i = 0; i < sw->nou; ++i) {
        s = sw->out[i];
        for (l = 0; s.prev != TRANSPOSIT_EXT; ++l) {
            if  (l > sw->num) {
                errno = EMLINK;
                ex = EX_DATAERR;
                goto out;
            }
            s = sw->net[s.prev].src[(sw->net[s.prev].state) ? (1 - s.lnkn) : s.lnkn]
            ;
        }
        swr->swt[i] = s.lnkn;
    }
out:if  (ex) {
        if  (f) free(swr);
        if  (swe) *swe = NULL;
        swr = NULL;
    }
    return(swr);
}

/*****************************************************************************************************
 *       2x2 sw   o.                                      *
 *     ,  ,   2x2  ,            *
 *  .                                                                               *
 *        ,  0 ,                  *
 *  1 ,   ( - 0,  - 1)                     *
 *     :  ,  .                                       *
 *     :,  *****:   .           *
 *****************************************************************************************************/
void
transposit_2swdump(transposit_2sw *sw, FILE *o) {
    size_t i, k;

    fprintf( o, "%%%%%% %08X in:%"BLIN_U" x [%"BLIN_U"] x out:%"BLIN_U"\n"
           , sw->flags
           , BLIN_I(sw->nin)
           , BLIN_I(sw->num)
           , BLIN_I(sw->nou)
           )
    ;
    for (i = 0; i < sw->num; ++i) {
        fprintf(o, "%5"BLIN_U"<:", BLIN_I(i));
        for (k = 0; k < 2; ++k) {
            if  (sw->net[i].src[k].prev == TRANSPOSIT_EXT) {
                fprintf(o, "  *****:%u", sw->net[i].src[k].lnkn);
            } else {
                fprintf(o, "  %5u:%u", sw->net[i].src[k].prev, sw->net[i].src[k].lnkn);
        }   }
        fprintf(o, "  > %u <\n", sw->net[i].state);
    }
    for (i = 0; i < sw->nou; ++i) {
        fprintf(o, "%5"BLIN_U">:", BLIN_I(i));
        if  (sw->out[i].prev == TRANSPOSIT_EXT) {
            fprintf(o, "  *****:%u\n", sw->out[i].lnkn);
        } else {
            fprintf(o, "  %5u:%u\n", sw->out[i].prev, sw->out[i].lnkn);
}   }   }
